from dynesty import plotting as dyplot

from autofit.plot import SamplesPlotter

from autofit.plot.samples_plotters import skip_plot_in_test_mode
from autofit.plot.samples_plotters import log_value_error

class DynestyPlotter(SamplesPlotter):

    @skip_plot_in_test_mode
    def boundplot(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``boundplot``.
        
        This figure plots the bounding distribution used to propose either (1) live points
        at a given iteration or (2) a specific dead point during
        the course of a run, projected onto the two dimensions specified
        by `dims`.
        """

        dyplot.boundplot(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="boundplot")
        self.close()

    @skip_plot_in_test_mode
    def cornerbound(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``cornerbound``.

        This figure plots the bounding distribution used to propose either (1) live points
        at a given iteration or (2) a specific dead point during
        the course of a run, projected onto all pairs of dimensions.
        """
        
        dyplot.cornerbound(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerbound")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def cornerplot(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``cornerplot``.

        This figure plots a corner plot of the 1-D and 2-D marginalized posteriors.
        """
        dyplot.cornerplot(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerplot")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def cornerpoints(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``cornerpoints``.

        This figure plots a (sub-)corner plot of (weighted) samples.
        """
        dyplot.cornerpoints(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerpoints")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def runplot(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``runplot``.

        This figure plots live points, ln(likelihood), ln(weight), and ln(evidence)
        as a function of ln(prior volume).
        """
        dyplot.runplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="runplot")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def traceplot(self, **kwargs):
        """
        Plots the in-built ``dynesty`` plot ``traceplot``.

        This figure plots traces and marginalized posteriors for each parameter.
        """
        dyplot.traceplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="traceplot")
        self.close()


import matplotlib.pyplot as plt
import numpy as np

from autofit.plot.samples_plotters import MCMCPlotter


class PySwarmsPlotter(MCMCPlotter):

    def contour(self, **kwargs):
        from pyswarms.utils import plotters
        plotters.plot_contour(
            pos_history=self.samples.points,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="contour")
        self.close()

    def cost_history(self, **kwargs):
        from pyswarms.utils import plotters
        plotters.plot_cost_history(
            cost_history=self.samples.log_posterior_list,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cost_history")
        self.close()

    def trajectories(self, **kwargs):

        points = self.samples.points

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7))

        for i in range(self.samples.model.prior_count):
            ax = axes[i]
            ax.plot(np.asarray(points)[:, -1, i], self.samples.log_posterior_list, "k", alpha=0.3)
            ax.set_ylabel("Log Likelihood")
            ax.set_xlabel(self.model.parameter_labels_with_superscripts_latex[i])

        self.output.to_figure(structure=None, auto_filename="trajectories")
        self.close()

    def time_series(self, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7), sharex=True)
        points = self.samples.points

        for i in range(self.samples.model.prior_count):
            ax = axes[i]
            ax.plot(np.asarray(points)[:, -1, i], "k", alpha=0.3)
            ax.set_ylabel(self.model.parameter_labels_with_superscripts_latex[i])

        axes[-1].set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="time_series")
        self.close()



from typing import Union, List

from autofit.graphical.declarative.factor.hierarchical import HierarchicalFactor
from autofit.graphical.expectation_propagation import EPHistory
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.factor_graphs.factor import Factor
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import Collection
from autofit.non_linear.result import Result, AbstractResult
from autofit.non_linear.samples.samples import Samples


class HierarchicalResult(AbstractResult):
    def __init__(
            self,
            results: List[AbstractResult]
    ):
        """
        One true factor is created per each variable drawn from the declarative
        hierarchical factor. One result for each of those factors is in this
        collection.

        Parameters
        ----------
        results
            Results from hierarchical factor optimisations
        """
        super().__init__(
            results[0].sigma
        )
        self.results = results

    @property
    def samples(self) -> Samples:
        """
        All samples from all underlying optimisations are summed together.
        """
        return sum(
            result.samples
            for result
            in self.results
        )

    @property
    def model(self) -> AbstractPriorModel:
        """
        Priors with the same path in the model are combined by taking the
        product of their underlying messages. i.e. the product of the
        posteriors.

        The distribution model is returned.
        """
        return AbstractPriorModel.product(
            result.model
            for result
            in self.results
        ).distribution_model

    @property
    def instance(self):
        """
        Return the instance (e.g. prior) describing the distribution
        from which samples are drawn.
        """
        return super().instance.distribution_model


class EPResult:
    def __init__(
            self,
            ep_history: EPHistory,
            declarative_factor,
            updated_ep_mean_field: EPMeanField,
    ):
        """
        The result of an EP Optimisation including its history, a declarative
        representation of the optimised graph and the resultant EPMeanField
        which comprises factors and resultant approximations of variables.

        Parameters
        ----------
        ep_history
            A history of the optimisation
        declarative_factor: AbstractDeclarativeFactor
            A declarative representation of the factor being optimised
        updated_ep_mean_field
            An updated mean field; effectively the result of the optimisation
        """
        self.ep_history = ep_history
        self.declarative_factor = declarative_factor
        self.updated_ep_mean_field = updated_ep_mean_field

    @property
    def model(self) -> Collection:
        """
        A collection populated with messages representing the posteriors of
        the EP Optimisation. Each item in the collection represents a single
        factor in the optimisation.
        """
        collection = Collection({
            factor.name: factor.prior_model
            for factor
            in self.declarative_factor.model_factors
        })
        arguments = {
            prior: prior.with_message(
                self.updated_ep_mean_field.mean_field[
                    prior
                ]
            )
            for prior
            in collection.priors
        }

        return collection.gaussian_prior_model_for_arguments(
            arguments
        )

    @property
    def latest_results(self) -> List[Result]:
        """
        A list of results for all analysis and hierarchical factors.
        """
        return [
            self.ep_history[factor].latest_result
            for factor in self.declarative_factor.model_factors
        ]

    @property
    def instance(self):
        """
        The median instance taken from the updated model
        """
        return self.model.instance_from_prior_medians()

    def latest_for(
            self,
            factor: Union[Factor, HierarchicalFactor]
    ) -> AbstractResult:
        """
        Return the latest result for a factor.

        If the factor is hierarchical return the latest result for
        the first true factor (i.e. for one of the drawn variable)

        Parameters
        ----------
        factor
            A factor from the graph

        Returns
        -------
        The latest result for that factor
        """
        if isinstance(factor, HierarchicalFactor):
            results = [
                self.ep_history[child_factor].latest_result
                for child_factor
                in factor.factors
            ]
            return HierarchicalResult(results)
        return self.ep_history[factor].latest_result



import json
import logging
import os
import re
import shutil
import zipfile
from abc import ABC, abstractmethod
from configparser import NoSectionError
from os import path
from typing import Optional

from autoconf import conf
from autofit.mapper.identifier import Identifier, IdentifierField
from autofit.text import text_util
from autofit.tools.util import open_, zip_directory

logger = logging.getLogger(__name__)

pattern = re.compile(r"(?<!^)(?=[A-Z])")


class AbstractPaths(ABC):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths=True,
        parent: Optional["AbstractPaths"] = None,
        unique_tag: Optional[str] = None,
        identifier: str = None,
    ):
        """
        Manages the path structure for `NonLinearSearch` output, for analyses both not using and using the search
        API. Use via non-linear searches requires manual input of paths, whereas the search API manages this using the
        search attributes.

        The output path within which the *Paths* objects path structure is contained is set via PyAutoConf, using the
        command:

        from autoconf import conf
        conf.instance = conf.Config(output_path="path/to/output")

        If we assume all the input strings above are used with the following example names:

        name = "name"
        path_prefix = "folder_0/folder_1"

        The output path of the `NonLinearSearch` results will be:

        /path/to/output/folder_0/folder_1/name

        Parameters
        ----------
        name
            The name of the non-linear search, which is used as a folder name after the ``path_prefix``. For searchs
            this name is the ``name``.
        path_prefix
            A prefixed path that appears after the output_path but before the name variable.
        is_identifier_in_paths
            If True output path and symlink path terminate with an identifier generated from the
            search and model
        """

        self.name = name or ""
        self.path_prefix = path_prefix or ""

        self.unique_tag = unique_tag

        self._non_linear_name = None
        self.__identifier = identifier or None

        self.is_identifier_in_paths = is_identifier_in_paths

        self._parent = None
        self.parent = parent

        try:
            self.remove_files = conf.instance["general"]["output"]["remove_files"]

            if conf.instance["general"]["hpc"]["hpc_mode"]:
                self.remove_files = True
        except NoSectionError as e:
            logger.exception(e)

    def save_parent_identifier(self):
        pass

    def save_unique_tag(self, is_grid_search=False):
        pass

    def __str__(self):
        return self.output_path

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    @property
    def parent(self) -> "AbstractPaths":
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.
        """
        return self._parent

    @parent.setter
    @abstractmethod
    def parent(self, parent: "AbstractPaths"):
        pass

    @property
    @abstractmethod
    def is_grid_search(self) -> bool:
        pass

    def for_sub_analysis(self, analysis_name: str):
        return self.create_child(name=analysis_name)

    @abstractmethod
    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "AbstractPaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory.

        Parameters
        ----------
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.
        identifier

        Returns
        -------
        A new paths object
        """

    search = IdentifierField()
    model = IdentifierField()
    unique_tag = IdentifierField()

    @abstractmethod
    def save_named_instance(self, name: str, instance):
        """
        Save an instance, such as that at a given sigma
        """

    @property
    def non_linear_name(self):
        if self._non_linear_name is None:
            if self.search is not None:
                self._non_linear_name = pattern.sub(
                    "_", type(self.search).__name__
                ).lower()
        return self._non_linear_name

    @property
    def _identifier(self):
        if self.__identifier is None:
            if None in (self.model, self.search):
                logger.debug(
                    "Generating identifier without both model and search having been set."
                )

            identifier_list = [self.search, self.model]

            if self.unique_tag is not None:
                identifier_list.append(self.unique_tag)
            self.__identifier = Identifier(identifier_list)

        return self.__identifier

    @_identifier.setter
    def _identifier(self, identifier):
        self.__identifier = identifier

    @property
    def identifier(self):
        return str(self._identifier)

    def save_identifier(self):
        with open_(f"{self.output_path}/.identifier", "w+") as f:
            f.write(self._identifier.description)

    @property
    def samples_path(self) -> str:
        """
        The path to the samples folder.
        """
        return path.join(self.output_path, "samples")

    @property
    def image_path(self) -> str:
        """
        The path to the image folder.
        """
        return path.join(self.output_path, "image")

    @property
    def profile_path(self) -> str:
        """
        The path to the profile folder.
        """
        return path.join(self.output_path, "profile")

    @property
    def output_path(self) -> str:
        """
        The path to the output information for a search.
        """
        strings = list(
            filter(len, [str(conf.instance.output_path), self.path_prefix, self.name,],)
        )

        if self.is_identifier_in_paths:
            strings.append(self.identifier)

        return path.join("", *strings)

    def zip_remove(self):
        """
        Copy files from the sym linked search folder then remove the sym linked folder.
        """

        self._zip()

    def _zip(self):

        try:
            zip_directory(self.output_path, self._zip_path)

            if self.remove_files:
                shutil.rmtree(
                    self.output_path, ignore_errors=True,
                )

        except FileNotFoundError:
            pass

    def zip_remove_nuclear(self):
        """
        When multiple model-fits are performed using the same `path_prefix` and `name`,
        the results are populated in the same folder with different unique identifiers.

        By accident, one may perform runs where additional results are placed
        in these folders which are not wanted for the subsequent analysis.

        Removing these results from the directory can be cumbersome, as determining
        the unwanted results based on their unique identifier requires visually inspecting
        them.

        These unwanted results can also make manipulating the results via the database
        problematic, as one may need to again filter based on unique identifier.

        When a run is performed in nuclear mode, all results in every folder are
        deleted except the results corresponding to the unique identifier of that run.

        Therefore, provided the user is 100% certain that the run corresponds to the
        results they want to keep, nuclear mode can be used to remove all unwanted results.

        For example, suppose a folder has 5 results, 4 of which are unwanted and 1 which is
        wanted. If nuclear mode runs, and the model-fit is set up correctly such that the
        identifier created corresponds to the wanted result, all 4 unwanted results
        will be deleted.

        To enable nuclear mode, set the environment variable ``PYAUTOFIT_NUCLEAR_MODE=1``.

        Nuclear mode is dangerous, and must be used with CAUTION AND CARE!
        """

        if os.environ.get("PYAUTOFIT_NUCLEAR_MODE") == "1":

            file_path = os.path.split(self.output_path)[0]

            file_list = os.listdir(file_path)
            file_list = [file for file in file_list if self.identifier not in file]

            for file in file_list:

                file_to_remove = path.join(file_path, file)

                try:
                    os.remove(file_to_remove)
                    logger.info(f"NUCLEAR MODE -- Removed {file_to_remove}")
                except (IsADirectoryError, FileNotFoundError):
                    pass

                try:
                    shutil.rmtree(file_to_remove)
                    logger.info(f"NUCLEAR MODE -- Removed {file_to_remove}")
                except (NotADirectoryError, FileNotFoundError):
                    pass

    def restore(self):
        """
        Copy files from the ``.zip`` file to the samples folder.
        """

        if path.exists(self._zip_path):
            shutil.rmtree(self.output_path, ignore_errors=True)

            try:
                with zipfile.ZipFile(self._zip_path, "r") as f:
                    f.extractall(self.output_path)
            except zipfile.BadZipFile as e:
                raise zipfile.BadZipFile(
                    f"Unable to restore the zip file at the path {self._zip_path}"
                ) from e

            os.remove(self._zip_path)

    def __eq__(self, other):
        return isinstance(other, AbstractPaths) and all(
            [
                self.path_prefix == other.path_prefix,
                self.name == other.name,
                self.non_linear_name == other.non_linear_name,
            ]
        )

    @property
    def _zip_path(self) -> str:
        return f"{self.output_path}.zip"

    @abstractmethod
    def save_object(self, name: str, obj: object):
        pass

    @abstractmethod
    def load_object(self, name: str):
        pass

    @abstractmethod
    def remove_object(self, name: str):
        pass

    @abstractmethod
    def is_object(self, name: str) -> bool:
        pass

    @property
    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def completed(self):
        pass

    @abstractmethod
    def save_all(self, search_config_dict=None, info=None, pickle_files=None):
        pass

    @abstractmethod
    def load_samples(self):
        """
        Load samples from the database
        """

    @abstractmethod
    def save_samples(self, samples):
        """
        Save samples to the database
        """

    def samples_to_csv(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """

    @abstractmethod
    def load_samples_info(self):
        pass

    def _save_search(self, config_dict):

        with open_(path.join(self.output_path, "search.json"), "w+") as f:
            json.dump(dict(config_dict), f, indent=4)

    def save_summary(self, samples, log_likelihood_function_time):

        result_info = text_util.result_info_from(samples=samples,)

        filename = path.join(self.output_path, "model.results")

        with open_(filename, "w") as f:
            f.write(result_info)

        text_util.search_summary_to_file(
            samples=samples,
            log_likelihood_function_time=log_likelihood_function_time,
            filename=path.join(self.output_path, "search.summary"),
        )

    @property
    def _samples_file(self) -> str:
        return path.join(self.samples_path, "samples.csv")

    @property
    def _info_file(self) -> str:
        return path.join(self.samples_path, "info.json")



from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from dynesty import NestedSampler as StaticSampler
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.nest.dynesty.samples import SamplesDynesty

from .abstract import AbstractDynesty, prior_transform


class DynestyStatic(AbstractDynesty):
    __identifier_fields__ = (
        "nlive",
        "bound",
        "sample",
        "bootstrap",
        "enlarge",
        "walks",
        "facc",
        "slices",
        "fmove",
        "max_move"
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer=None,
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        A Dynesty `NonLinearSearch` using a static number of live points.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        iterations_per_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating DynestyStatic Search")

    def samples_from(self, model):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For Dynesty, all information that we need is available from the instance of the dynesty sampler.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """
        sampler = StaticSampler.restore(self.checkpoint_file)

        return SamplesDynesty.from_results_internal(
            model=model,
            results_internal=sampler.results,
            number_live_points=self.total_live_points,
            unconverged_sample_size=1,
            time=self.timer.time,
        )

    def sampler_from(
            self,
            model: AbstractPriorModel,
            fitness_function,
            checkpoint_exists : bool,
            pool: Optional,
            queue_size: Optional[int]
    ):
        """
        Returns an instance of the Dynesty static sampler set up using the input variables of this class.

        If no existing dynesty sampler exist on hard-disk (located via a `checkpoint_file`) a new instance is
        created with which sampler is performed. If one does exist, the dynesty `restore()` function is used to
        create the instance of the sampler.

        Dynesty samplers with a multiprocessing pool may be created by inputting a dynesty `Pool` object, however
        non pooled instances can also be created by passing `pool=None` and `queue_size=None`.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space.
        fitness_function
            An instance of the fitness class used to evaluate the likelihood of each model.
        pool
            A dynesty Pool object which performs likelihood evaluations over multiple CPUs.
        queue_size
            The number of CPU's over which multiprocessing is performed, determining how many samples are stored
            in the dynesty queue for samples.
        """

        if checkpoint_exists:

            sampler = StaticSampler.restore(
                fname=self.checkpoint_file,
                pool=pool
            )

            uses_pool = self.read_uses_pool()

            self.check_pool(uses_pool=uses_pool, pool=pool)

            return sampler

        else:

            live_points = self.live_points_init_from(model=model, fitness_function=fitness_function)

            if pool is not None:

                self.write_uses_pool(uses_pool=True)

                return StaticSampler(
                    loglikelihood=pool.loglike,
                    prior_transform=pool.prior_transform,
                    ndim=model.prior_count,
                    live_points=live_points,
                    queue_size=queue_size,
                    pool=pool,
                    **self.config_dict_search
                )

            self.write_uses_pool(uses_pool=False)

            return StaticSampler(
                loglikelihood=fitness_function,
                prior_transform=prior_transform,
                ndim=model.prior_count,
                logl_args=[model, fitness_function],
                ptform_args=[model],
                live_points=live_points,
                **self.config_dict_search
            )

    @property
    def total_live_points(self):
        return self.config_dict_search["nlive"]
